"""Downloads all the data necessary to run ECCO_advect_2D.py and ECCO_advect_3D.py"""
import sys
from pathlib import Path

sys.path.append(str(Path(__file__).parent))

from data_downloaders.download_and_process_currents import download_and_process_currents
from data_downloaders.download_10m_wind import download_and_interpolate_ncep_ncar_wind
from data_downloaders.get_ECCO_credentials import get_ECCO_credentials
from data_downloaders.download_and_process_density import download_and_process_density


def main():
    data_root = Path(input("Input directory to download data into: "))
    data_root.mkdir(exist_ok=True)
    print("")
    print("Downloading ncep-ncar reanalysis ii 10m wind...")
    download_and_interpolate_ncep_ncar_wind(out_dir=data_root)

    print("")
    print("Downloading Seawater Density from JPL's ECCOv4r4...")
    density_path = data_root / "RHO_2015.nc"
    if density_path.exists():
        print(f"{density_path} already exists.  Skipping...")
    else:
        user, password = get_ECCO_credentials()
        download_and_process_density(
            out_path=density_path, user=user, password=password
        )

    print("")
    print("Downloading 3D currents from JPL's ECCOv4r4...")
    download_and_process_currents(out_dir=data_root)

    print("")
    print("All done!")


if __name__ == "__main__":
    main()



"""
This file contains constants which are defined within the kernel runtime, hard-copied here in order to make them
accessible to the python runtime.
"""

# this is defined by the OpenCL runtime itself; this matches the definition in the OpenCL 1.2 standard.
UINT_MAX = 0xFFFFFFFF

# these match the definitions in src/kernels/exit_codes.cl.
EXIT_CODES = {
    0: "SUCCESS",
    1: "NULL_LOCATION",
    2: "INVALID_LATITUDE",
    3: "SEAWATER_DENSITY_LOOKUP_FAILURE",
    -1: "INVALID_ADVECTION_SCHEME",
}



from datetime import datetime
from pathlib import Path
from typing import Tuple

import numpy as np
import pandas as pd
import xarray as xr


def generate_3D_sourcefile(
    num_particles: int,
    density_range: Tuple[float, float],
    radius_range: Tuple[float, float],
    corey_shape_factor_range: Tuple[float, float],
    depth_range: Tuple[float, float],
    release_date_range: Tuple[datetime, datetime],
    out_path: Path,
    land_mask_path: str = Path(__file__).parent / "land_mask.nc",
):
    """
    generates a sourcefile with uniformly random properties within given bounds
    radius is distributed uniformly in log space
    :param num_particles: number of particles to generate
    :param density_range: density of particles, kg m^-3
    :param radius_range: radius of particles, m
    :param corey_shape_factor_range: unitless representation of 3d shape.  Domain: (.15, 1]
    :param depth_range: starting depth of particles, m (positive up)
    :param release_date_range: timestamp at which particle is released
    :param out_path: path at which to save file
    :param land_mask_path: path which contains a boolean land mask netcdf, dimensions {lon, lat}
    :return:
    """
    ds = create_2D_source_dataset(
        num_particles=num_particles,
        release_date_range=release_date_range,
        land_mask_path=land_mask_path,
    )
    rng = np.random.default_rng()
    ds = ds.assign(
        {
            "depth": ("p_id", rng.uniform(*depth_range, num_particles)),
            "radius": (
                "p_id",
                10 ** rng.uniform(*np.log10(radius_range), num_particles),
            ),
            "density": ("p_id", rng.uniform(*density_range, num_particles)),
            "corey_shape_factor": (
                "p_id",
                rng.uniform(*corey_shape_factor_range, num_particles),
            ),
        }
    )
    ds.attrs["title"] = f"3D Sourcefile for ADVECTOR"
    ds.to_netcdf(out_path)


def generate_2D_sourcefile(
    num_particles: int,
    release_date_range: Tuple[datetime, datetime],
    out_path: str,
    land_mask_path: str = Path(__file__).parent / "land_mask.nc",
):
    """
    generates a sourcefile with uniformly random properties within given bounds
    :param num_particles: number of particles to generate
    :param release_date_range: timestamp at which particle is released
    :param out_path: path at which to save file
    :param land_mask_path: path which contains a boolean land mask netcdf, dimensions {lon, lat}
    :return:
    """
    ds = create_2D_source_dataset(
        num_particles=num_particles,
        release_date_range=release_date_range,
        land_mask_path=land_mask_path,
    )
    ds.attrs["title"] = f"2D Sourcefile for ADVECTOR"
    ds.to_netcdf(out_path)


def create_2D_source_dataset(
    num_particles: int,
    release_date_range: Tuple[datetime, datetime],
    land_mask_path: str = Path(__file__).parent / "land_mask.nc",
):
    rng = np.random.default_rng()
    # create a land mask
    land = xr.open_dataarray(land_mask_path)

    # initialize particles
    [X, Y] = np.meshgrid(land.lon, land.lat)
    ocean_points = np.array([X[~land], Y[~land]]).T
    p0 = pd.DataFrame(
        data=rng.choice(ocean_points, size=num_particles, replace=True),
        columns=["lon", "lat"],
    )
    p0["release_date"] = pd.to_datetime(
        rng.uniform(
            release_date_range[0].timestamp(),
            release_date_range[1].timestamp(),
            num_particles,
        ),
        unit="s",
    )
    p0["p_id"] = np.arange(num_particles)
    ds = xr.Dataset(p0.set_index("p_id"))
    ds.attrs["institution"] = "The Ocean Cleanup"
    return ds



"""
advect on ECCO currents
"""
from datetime import datetime, timedelta
from pathlib import Path

import xarray as xr

from ADVECTOR.examples.helpers.generate_sourcefiles import generate_3D_sourcefile
from ADVECTOR.examples.helpers.generate_configfile import generate_sample_configfile
from ADVECTOR.plotting.plot_advection import animate_ocean_advection
from ADVECTOR.run_advector_3D import run_advector_3D


def main():
    data_root = Path(input("Input path to example data directory: "))
    output_root = Path(input("Input path to directory for outputfiles: "))
    output_root.mkdir(exist_ok=True)

    ADVECTION_START = datetime(2015, 1, 1)
    ADVECTION_END = datetime(2015, 2, 1)

    # generate a sourcefile
    sourcefile_path = output_root / "3D_uniform_source_2015.nc"
    generate_3D_sourcefile(
        num_particles=5000,
        density_range=(800, 1000),
        radius_range=(1e-6, 1),
        corey_shape_factor_range=(0.15, 1),
        release_date_range=(
            ADVECTION_START,
            ADVECTION_START + (ADVECTION_END - ADVECTION_START) / 2,
        ),
        depth_range=(0, 0),
        out_path=sourcefile_path,
    )
    # generate a configfile
    configfile_path = output_root / "config.nc"
    generate_sample_configfile(out_path=configfile_path)

    def preprocess_currents(currents: xr.Dataset) -> xr.Dataset:
        return currents.rename(
            {
                "longitude": "lon",
                "latitude": "lat",
                "Z": "depth",
                "EVEL": "U",
                "NVEL": "V",
                "WVELMASS": "W",
            }
        )

    def preprocess_wind(wind: xr.Dataset) -> xr.Dataset:
        return wind.rename({"uwnd": "U", "vwnd": "V", "level": "depth"})

    out_dir = output_root / sourcefile_path.stem
    out_dir.mkdir(parents=True, exist_ok=True)

    # run the model!
    out_paths = run_advector_3D(
        output_directory=str(out_dir),
        sourcefile_path=str(sourcefile_path),
        configfile_path=str(configfile_path),
        u_water_path=str(data_root / "EVEL_2015*.nc"),
        v_water_path=str(data_root / "NVEL_2015*.nc"),
        w_water_path=str(data_root / "WVELMASS_2015*.nc"),
        water_preprocessor=preprocess_currents,
        u_wind_path=str(data_root / "uwnd.10m.gauss.2015.nc"),
        v_wind_path=str(data_root / "vwnd.10m.gauss.2015.nc"),
        wind_preprocessor=preprocess_wind,
        seawater_density_path=str(data_root / "RHO_2015.nc"),
        advection_start_date=ADVECTION_START,
        timestep=timedelta(hours=1),
        num_timesteps=24 * (ADVECTION_END - ADVECTION_START).days,
        save_period=4,
        memory_utilization=0.4,
        overwrite_existing_files=True,
    )

    for out_path in out_paths:
        print("Animating trajectories...")
        animate_ocean_advection(out_path)


if __name__ == "__main__":
    main()



from enum import Enum

import matplotlib.pyplot as plt
import numpy as np
import xarray as xr
from matplotlib import colors

from ..io_tools.create_bathymetry import create_bathymetry_from_land_mask


class BathyPlotType(Enum):
    gridded = 0
    contour = 1


def plot_3d_trajectories(
    particles: xr.Dataset,
    land_mask: xr.DataArray,
    bathymetry_plot_type: BathyPlotType = BathyPlotType.gridded,
):
    plt.figure(figsize=(12, 8))
    ax = plt.axes(projection="3d")
    ax.view_init(15, -80)
    if bathymetry_plot_type == BathyPlotType.gridded:
        plot_gridded_bathymetry(ax=ax, land_mask=land_mask)
    elif bathymetry_plot_type == BathyPlotType.contour:
        plot_contour_bathymetry(ax=ax, land_mask=land_mask)
    else:
        raise ValueError("Invalid BathyPlotType")

    for i in range(len(particles.p_id)):
        ax.scatter3D(
            particles.isel(p_id=i).lon,
            particles.isel(p_id=i).lat,
            particles.isel(p_id=i).depth,
            ".",
            s=3,
        )
    ax.set_xlim([land_mask.lon.min(), land_mask.lon.max()])
    ax.set_ylim([land_mask.lat.min(), land_mask.lat.max()])
    ax.set_zlim([min(land_mask.depth), -1 * (min(land_mask.depth))])
    ax.set_xlabel("Longitude (deg E)")
    ax.set_ylabel("Latitude (deg N)")
    ax.set_zlabel("Depth (m)")


def plot_gridded_bathymetry(
    ax: plt.Axes,
    land_mask: xr.DataArray,
):
    bathy = create_bathymetry_from_land_mask(land_mask=land_mask)
    floor = np.min(bathy.values) - 1
    X, Y = np.meshgrid(bathy.lon, bathy.lat)
    dx = np.diff(bathy.lon)[0]
    dy = np.diff(bathy.lat)[0]
    facecolors = np.empty((len(X.ravel()), 6), dtype=object)
    facecolors[:] = "tab:blue"
    facecolors[bathy.values.ravel() == 0, 1] = "tab:green"

    ax.bar3d(
        X.ravel(),
        Y.ravel(),
        floor,
        dx,
        dy,
        bathy.values.ravel() - floor,
        edgecolor="k",
        linewidth=0.1,
        alpha=0.8,
        color=facecolors.ravel(),
    )


def plot_contour_bathymetry(
    ax: plt.Axes,
    land_mask: xr.DataArray,
):
    bathy = create_bathymetry_from_land_mask(land_mask=land_mask)
    X, Y = np.meshgrid(bathy.lon, bathy.lat)
    color = np.zeros_like(bathy)
    color[bathy.values == 0] = 1
    cmap = colors.ListedColormap(["tab:blue", "tab:green"])

    my_col = cmap(color)
    ax.plot_surface(X=X, Y=Y, Z=bathy.values, alpha=0.9, facecolors=my_col)


if __name__ == "__main__":
    # example plotting around cape horn
    lon_range = (-80, -50)
    lat_range = (-60, -40)
    depth_range = (-10000, 0)
    P = xr.open_dataset(
        "../../examples/outputfiles/ECCO_2015_3D/3D_uniform_source_2015/ADVECTOR_3D_output_2015.nc",
    )
    P = P.isel(
        p_id=(
            (P.lon >= lon_range[0])
            & (P.lon <= lon_range[1])
            & (P.lat >= lat_range[0])
            & (P.lat <= lat_range[1])
            & (P.depth >= depth_range[0])
            & (P.depth <= depth_range[1])
        ).any(dim="time")
    )
    ECCO_land_mask = (
        xr.open_dataset("../../examples/ECCO/currents/U_2015-01-01.nc")
        .squeeze()
        .U.isnull()
        .sortby("depth")
        .sel(
            lon=slice(*lon_range),
            lat=slice(*lat_range),
        )
    )
    plot_3d_trajectories(
        particles=P.isel(p_id=0),
        land_mask=ECCO_land_mask,
        bathymetry_plot_type=BathyPlotType.gridded,
    )



from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Dict

import numpy as np
import pandas as pd
import pyopencl as cl
import xarray as xr

from ..enums.forcings import Forcing


@dataclass
class KernelConfig(ABC):
    pass


class Kernel(ABC):
    def __init__(
        self,
        forcing_data: Dict[Forcing, xr.Dataset],
        p0: xr.Dataset,
        advect_time: pd.DatetimeIndex,
        save_every: int,
        config: KernelConfig,
        context: cl.Context,
    ):
        """
        :param forcing_data: dict containing forcing datasets
        :param p0: initial state of particles
        :param advect_time: the timeseries which the kernel advects on
        :param save_every: number of timesteps between each writing of particle state
        :param config: dataclass with any extra settings needed by subclass implementation
        :param context: PyopenCL context for executing OpenCL programs
        """
        # save some arguments for creating output dataset
        self.p0 = p0
        self.out_time = advect_time[::save_every][1:]
        self.advect_time = advect_time

        # advection time parameters
        self.start_time = np.float64(advect_time[0].timestamp())
        self.dt = np.float64(pd.Timedelta(advect_time.freq).total_seconds())
        self.ntimesteps = np.uint32(len(advect_time) - 1)  # initial position given
        self.save_every = np.uint32(save_every)

        # create opencl objects
        self.context = context
        self.queue = cl.CommandQueue(context)
        self.cl_kernel = cl.Program(context, self._kernel_source_code).build(
            options=["-I", str(self._model_core_path)]
        )

        # some handy timers
        self.data_load_time = 0
        self.buf_time = 0
        self.kernel_time = 0

    @property
    def _model_core_path(self):
        return Path(__file__).parent.parent / "model_core"

    @property
    @abstractmethod
    def _kernel_source_code(self) -> str:
        pass

    @abstractmethod
    def execute(self) -> xr.Dataset:
        pass

    @abstractmethod
    def get_memory_footprint(self) -> dict:
        pass

    @abstractmethod
    def get_data_loading_time(self) -> float:
        pass

    @abstractmethod
    def get_buffer_transfer_time(self) -> float:
        pass

    @abstractmethod
    def get_kernel_execution_time(self) -> float:
        pass



"""
simple script to download daily-mean 10m wind from the ncep ncar doe ii reanalysis
"""
import subprocess
import tempfile
from pathlib import Path

import numpy as np
import xarray as xr


def download_and_interpolate_ncep_ncar_wind(out_dir: Path):
    with tempfile.TemporaryDirectory() as temp_dir:
        urls = [
            "ftp://ftp2.psl.noaa.gov/Datasets/ncep.reanalysis2.dailyavgs/gaussian_grid/uwnd.10m.gauss.2015.nc",
            "ftp://ftp2.psl.noaa.gov/Datasets/ncep.reanalysis2.dailyavgs/gaussian_grid/vwnd.10m.gauss.2015.nc",
        ]
        filenames = [url.split("/")[-1] for url in urls]

        for filename, url in zip(filenames, urls):
            out_path = out_dir / filename
            if out_path.exists():
                print(f"skipping {out_path}, already exists")
                continue
            subprocess.run(
                [
                    "wget",
                    "--retry-connrefused",
                    "--waitretry=1",
                    "--read-timeout=20",
                    "--timeout=15",
                    "-t 0",
                    "-q",
                    "--show-progress",
                    "-O",
                    str(Path(temp_dir) / filename),
                    url,
                ]
            )

            # interpolate to regular grids
            print(f"Interpolating latitude to evenly spaced grid for file {filename}")
            ds = xr.open_dataset(Path(temp_dir) / filename)
            ds = ds.sortby("lat", ascending=True)
            ds = ds.interp(lat=np.linspace(ds.lat.min(), ds.lat.max(), len(ds.lat)))
            ds.to_netcdf(out_path)



from pathlib import Path

import numpy as np
import xarray as xr


def generate_configfile(
    out_path: Path,
    horizontal_diffusivity: np.ndarray,
    z_hd: np.ndarray,
    vertical_diffusivity: np.ndarray,
    z_vd: np.ndarray,
    max_wave_height: float = None,
    wave_mixing_depth_factor: float = None,
):
    """
    script to generate a configuration file given requisite parameters
    """
    config = xr.Dataset(
        {
            "horizontal_diffusivity": (
                "z_hd",
                horizontal_diffusivity,
                {"units": "m^2 s^-1"},
            ),
            "vertical_diffusivity": (
                "z_vd",
                vertical_diffusivity,
                {"units": "m^2 s^-1"},
            ),
        },
        coords={
            "z_hd": (
                "z_hd",
                z_hd,
                {
                    "long_name": "depth coordinate for horizontal_diffusivity",
                    "units": "m",
                    "positive": "up",
                },
            ),
            "z_vd": (
                "z_vd",
                z_vd,
                {
                    "long_name": "depth coordinate for vertical_diffusivity",
                    "units": "m",
                    "positive": "up",
                },
            ),
        },
        attrs={
            "title": f"Configuration File for ADVECTOR",
            "institution": "The Ocean Cleanup",
        },
    )

    if max_wave_height:
        config["max_wave_height"] = max_wave_height
    if wave_mixing_depth_factor:
        config["wave_mixing_depth_factor"] = wave_mixing_depth_factor

    config.to_netcdf(out_path)


def generate_sample_configfile(out_path: Path):
    """a sample configuration file, diffusivity profiles are NOT based on true ocean state"""
    generate_configfile(
        horizontal_diffusivity=np.linspace(1500, 1, 20),  # m^2 s^-1
        z_hd=-np.logspace(0, 4, 20),  # m
        vertical_diffusivity=np.linspace(-5e-3, 1e-2, 10) ** 2,
        z_vd=np.linspace(-1e4, 0, 10),  # m
        out_path=out_path,
    )


if __name__ == "__main__":
    generate_sample_configfile(out_path=Path(__file__).parent / "sample_config.nc")



"""
Utilities which help split up advection based on memory constraints
"""
import math
from typing import Dict, List, Tuple

import numpy as np
import pandas as pd
import xarray as xr
from tqdm import tqdm

from ..enums.forcings import Forcing


def chunk_advection_params(
    device_bytes: int,
    forcing_data: Dict[Forcing, xr.Dataset],
    num_particles: int,
    advect_time: pd.DatetimeIndex,
    save_every: int,
) -> Tuple[List[pd.DatetimeIndex], List[Dict[Forcing, xr.Dataset]]]:
    """given the parameters for advection, return parameters for an iterative advection"""
    out_time = advect_time[::save_every]
    # each element of out_time marks a time at which the driver will return particle position

    # estimate total size of memory we need to eventually run through the device
    field_bytes, output_bytes, p0_bytes = estimate_memory_bytes(
        forcing_data=forcing_data,
        num_particles=num_particles,
        out_timesteps=len(out_time) - 1,
    )

    # it's hard to pre-compute the exact number of chunks necessary that will fit into RAM.  we start with a guess,
    # then if any of the chunks we create don't fit into RAM, we just increment the number of chunks and try again.
    print("\tIncrementing number of chunks until data fits...")
    num_chunks = math.ceil((field_bytes + output_bytes) / (device_bytes - p0_bytes))
    with tqdm(total=len(out_time) - num_chunks) as pbar:
        while True:
            if num_chunks > len(out_time):
                raise RuntimeError(
                    "There is not enough memory to hold even the smallest possible chunk!"
                )
            # now we split up the advection OUTPUT into chunks.  All else will be based on this splitting.
            out_time_chunks = np.array_split(out_time, num_chunks)
            # give subsequent time chunks overlapping endpoints.  This is because the final reported value
            # from a computation will be fed to the next computation as the start point.
            out_time_chunks[1:] = [
                out_time_chunks[i - 1][-1:].append(out_time_chunks[i])
                for i in range(1, len(out_time_chunks))
            ]

            # now, we split up the datasets.  If any chunk is too big, we
            # immediately quit, increment num_chunks, and try again.
            advect_time_chunks = []
            forcing_data_chunks = []
            all_chunks_fit = True
            for out_time_chunk in out_time_chunks:
                # extract dataset chunks for this out_time chunk
                advect_time_chunks.append(
                    advect_time[
                        (out_time_chunk[0] <= advect_time)
                        & (advect_time <= out_time_chunk[-1])
                    ]
                )
                forcing_data_chunks.append(
                    {
                        forcing: extract_dataset_chunk(ds, out_time_chunk)
                        for forcing, ds in forcing_data.items()
                    }
                )

                # if this chunk is too big, try again with more chunks.
                memory_bytes = estimate_memory_bytes(
                    forcing_data=forcing_data_chunks[-1],
                    num_particles=num_particles,
                    out_timesteps=len(out_time_chunk) - 1,
                )
                if sum(memory_bytes) > device_bytes:
                    all_chunks_fit = False
                    break
            if all_chunks_fit:
                return advect_time_chunks, forcing_data_chunks
            num_chunks += 1
            pbar.update(1)


def estimate_memory_bytes(
    forcing_data: Dict[Forcing, xr.Dataset],
    num_particles: int,
    out_timesteps: int,
) -> Tuple[int, int, int]:
    """This estimates total memory needed for the buffers.
    There's a bit more needed for the scalar arguments, but this is tiny"""
    forcing_data_bytes = 0
    for ds in forcing_data.values():
        field_shape = list(ds.data_vars.values())[0].shape
        # when passed to the kernel, field variables are represented as float32, coordinates as float64
        forcing_data_bytes += len(ds.data_vars) * 4 * math.prod(field_shape) + 8 * sum(
            field_shape
        )

    # output has three 4-byte variables for each particle per timestep, plus 1-byte exit codes
    output_bytes = 3 * 4 * num_particles * out_timesteps + 1 * num_particles
    # the particle input has 3 4-byte variables and 4 8-byte variables
    p0_bytes = (3 * 4 + 4 * 8) * num_particles
    return forcing_data_bytes, output_bytes, p0_bytes


def extract_dataset_chunk(
    dataset: xr.Dataset, out_time: pd.DatetimeIndex
) -> xr.Dataset:
    """
    Returns a list of slices from dataset, corresponding to the data needed for each chunk in out_time_chunks
    :param dataset: xarray Dataset with a "time" coordinate
    :param out_time: sequence of timestamps which define the computation domain;
        we will use their extent to slice dataset to match their time domain.
    :return:
    """
    start_time = dataset.time.sel(time=out_time[0], method="nearest")
    end_time = dataset.time.sel(time=out_time[-1], method="nearest")
    return dataset.sel(
        time=slice(start_time, end_time)
    )  # this slice is guaranteed to be at least length 1



from collections import OrderedDict
from typing import List

import numpy as np
import pyopencl as cl
import xarray as xr

from ..kernel_wrappers import kernel_constants


def is_uniformly_spaced_ascending(arr):
    tol = 1e-3
    return len(arr) == 1 or all(np.abs(np.diff(arr) - np.diff(arr)[0]) < tol)


def is_sorted_ascending(arr):
    return np.all(np.diff(arr) > 0)


def buffer_from_array(arr: np.ndarray, context: cl.Context) -> cl.Buffer:
    return cl.Buffer(
        context,
        cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR,
        hostbuf=arr.ravel(),
    )


class Field3D:
    """
    python wrapping of field3d, from src/model_core/fields.h.
    Enforces correct datatypes, provides convenience functions for kernel argument creation, etc.
    """

    def __init__(
        self,
        ds: xr.Dataset,
        varnames: List[str],
        non_uniform_time: bool = False,
    ):
        """
        :param ds: xarray dataset with some variables with coordinates {"time", "depth" (optional), "lat", "lon"},
        :param varnames: ordered list of variable names in ds (max 4 [U, V, W, bathy])
        """
        if len(varnames) < 1 or len(varnames) > 4:
            raise ValueError("Field3D must have between 1-4 variables")

        # create coordinate arrays, correct datatype, enforced spacing/sorting requirements, persisted into RAM
        self.coords = OrderedDict()
        self.coords["x"] = ds.lon.values.astype(np.float64)
        assert max(self.coords["x"]) <= 180
        assert min(self.coords["x"]) >= -180
        assert 1 <= len(self.coords["x"]) <= kernel_constants.UINT_MAX + 1
        assert is_uniformly_spaced_ascending(self.coords["x"])
        self.coords["y"] = ds.lat.values.astype(np.float64)
        assert max(self.coords["y"]) <= 90
        assert min(self.coords["y"]) >= -90
        assert 1 <= len(self.coords["y"]) <= kernel_constants.UINT_MAX + 1
        assert is_uniformly_spaced_ascending(self.coords["y"])
        if "depth" in ds.dims:
            self.coords["z"] = ds.depth.values.astype(np.float64)
            assert max(self.coords["z"]) <= 0
            assert is_sorted_ascending(self.coords["z"])
        # float64 representation of unix timestamp
        self.coords["t"] = ds.time.values.astype("datetime64[s]").astype(np.float64)
        assert 1 <= len(self.coords["t"]) <= kernel_constants.UINT_MAX + 1
        if non_uniform_time:
            assert is_sorted_ascending(self.coords["t"])
        else:
            assert is_uniformly_spaced_ascending(self.coords["t"])

        # put variables in an ordered dict, correct datatype, persisted into RAM
        if "z" in self.coords:
            ds = ds.transpose("time", "depth", "lat", "lon")
        else:
            ds = ds.transpose("time", "lat", "lon")
        self.variables = OrderedDict(
            (
                var,
                ds[var].astype(np.float32).values,
            )
            for var in varnames
        )

    def create_kernel_arguments(self, context: cl.Context) -> list:
        args = []
        for coord in self.coords.values():
            args.append(buffer_from_array(coord, context))
            args.append(np.uint32(len(coord)))
        for var in self.variables.values():
            args.append(buffer_from_array(var, context))
        return args

    def memory_usage_bytes(self) -> int:
        return sum(coord.nbytes for coord in self.coords.values()) + sum(
            var.nbytes for var in self.variables.values()
        )


def create_empty_2d_field():
    dummy_var = (["time", "lat", "lon"], [[[0]]])
    return Field3D(
        ds=xr.Dataset(
            data_vars={"U": dummy_var, "V": dummy_var},
            coords={"time": [0], "lat": [0], "lon": [0]},
        ),
        varnames=["U", "V"],
    )



import subprocess
import sys
from pathlib import Path

import matplotlib.animation as manimation
import matplotlib.cm as cm
import matplotlib.colors as mcol
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr
from tqdm import tqdm


def plot_ocean_trajectories(outputfile_path: str):
    """
    :param outputfile_path: must be a 2D outputfile
    """
    land_mask = xr.open_dataarray(outputfile_path, group="model_domain")
    fig, ax = plt.subplots(figsize=[14, 8])

    plot_grid(land_mask=land_mask, ax=ax)

    # plot trajectories
    P = xr.open_dataset(outputfile_path)
    ax.plot(P.lon.transpose("time", "p_id"), P.lat.transpose("time", "p_id"), ".")
    plt.show()


def animate_ocean_advection(
    outputfile_path: str,
    save: bool = False,
    colorbar_depth=None,
):
    P = xr.open_dataset(outputfile_path)
    fig, ax = plt.subplots(figsize=[14, 8])

    # plot the land mask
    if "depth" in P.variables:
        land_mask = xr.open_dataarray(outputfile_path, group="model_domain") == 0
    else:
        land_mask = xr.open_dataarray(outputfile_path, group="model_domain")
    plot_grid(land_mask=land_mask, ax=ax)

    # initialize the scatter plot with dummy data.
    trunc_winter = mcol.ListedColormap(cm.winter(np.linspace(0, 0.8, 100)))
    if "depth" in P.variables:
        vmin = colorbar_depth if colorbar_depth is not None else P.depth.min()
        dot = ax.scatter(
            np.zeros(len(P.p_id)),
            np.zeros(len(P.p_id)),
            c=np.zeros(len(P.p_id)),
            cmap=trunc_winter,
            s=5,
            norm=mcol.Normalize(vmin=vmin, vmax=0),
        )
        cbar = plt.colorbar(mappable=dot, ax=ax)
        cbar.ax.set_ylabel("Depth (m)")
    else:
        dot = ax.scatter(
            np.zeros(len(P.p_id)), np.zeros(len(P.p_id)), c="tab:blue", s=5
        )

    def base_update(i, P, ax, dot, timestr):
        dot.set_offsets(np.c_[np.array([P.isel(time=i).lon, P.isel(time=i).lat]).T])
        ax.set_title(timestr[i])
        ax.set_ylim(-90, 90)

    if "depth" in P.variables:

        def update_func(i, P, ax, dot, timestr):
            base_update(i, P, ax, dot, timestr)
            dot.set_array(P.isel(time=i).depth.values)

    else:
        update_func = base_update

    timestr = P.time.dt.strftime("%Y-%m-%dT%H:%M:%S").values
    if save:
        animate_ocean_advection_to_disk(
            outputfile_path, P, fig, ax, dot, timestr, update_func
        )
    else:
        for i in range(len(P.time)):
            update_func(i, P, ax, dot, timestr)
            plt.pause(0.005)
        plt.show()


def animate_ocean_advection_to_disk(
    outputfile_path, P, fig, ax, dot, timestr, update_func
):
    FFMpegWriter = manimation.writers["ffmpeg"]
    writer = FFMpegWriter(fps=30)
    outfile = Path(outputfile_path).with_suffix(".mp4")
    print("Creating Movie...")
    with writer.saving(fig, outfile=outfile, dpi=150):
        for i in tqdm(range(len(P.time))):
            update_func(i, P, ax, dot, timestr)
            writer.grab_frame()

    plt.close()
    print(f"Movie saved at {outfile}.")
    if sys.platform == "darwin":
        print("Opening Movie...")
        subprocess.call(["open", outfile])  # this won't work except on mac


def plot_grid(land_mask: xr.Dataset, ax: plt.Axes):
    xspacing = np.diff(land_mask.lon).mean()
    yspacing = np.diff(land_mask.lat).mean()
    lon_edges = np.append(
        (land_mask.lon - xspacing / 2), land_mask.lon[-1] + xspacing / 2
    )
    lat_edges = np.append(
        (land_mask.lat - yspacing / 2), land_mask.lat[-1] + yspacing / 2
    )
    ax.pcolormesh(lon_edges, lat_edges, ~land_mask, cmap="gray")



import os
from pathlib import Path

import pyopencl as cl

os.environ["PYOPENCL_COMPILER_OUTPUT"] = "1"

ROOT_DIR = Path(__file__).parent.parent
MODEL_CORE_DIR = ROOT_DIR / "ADVECTOR/model_core"
CL_CONTEXT = cl.create_some_context()
CL_QUEUE = cl.CommandQueue(CL_CONTEXT)



"""
advect on ECCO surface currents
"""
from datetime import datetime, timedelta
from pathlib import Path

import xarray as xr

from ADVECTOR.examples.helpers.generate_sourcefiles import generate_2D_sourcefile
from ADVECTOR.run_advector_2D import run_advector_2D
from ADVECTOR.plotting.plot_advection import (
    animate_ocean_advection,
    plot_ocean_trajectories,
)


def main():
    data_root = Path(input("Input path to example data directory: "))
    output_root = Path(input("Input path to directory for outputfiles: "))
    output_root.mkdir(exist_ok=True)

    ADVECTION_START = datetime(2015, 1, 1)
    ADVECTION_END = datetime(2015, 2, 1)

    sourcefile_path = output_root / "2D_uniform_source_2015.nc"
    generate_2D_sourcefile(
        num_particles=5000,
        release_date_range=(
            ADVECTION_START,
            ADVECTION_START + (ADVECTION_END - ADVECTION_START) / 2,
        ),
        out_path=sourcefile_path,
    )

    out_dir = output_root / sourcefile_path.stem
    out_dir.mkdir(parents=True, exist_ok=True)

    def preprocess_currents(currents: xr.Dataset) -> xr.Dataset:
        return currents.rename(
            {
                "longitude": "lon",
                "latitude": "lat",
                "Z": "depth",
                "EVEL": "U",
                "NVEL": "V",
            }
        )

    def preprocess_wind(wind: xr.Dataset) -> xr.Dataset:
        return wind.rename({"uwnd": "U", "vwnd": "V", "level": "depth"})

    out_paths = run_advector_2D(
        output_directory=str(out_dir),
        sourcefile_path=str(sourcefile_path),
        u_water_path=str(data_root / "EVEL_2015_01*.nc"),
        v_water_path=str(data_root / "NVEL_2015_01*.nc"),
        water_preprocessor=preprocess_currents,
        u_wind_path=str(data_root / "uwnd.10m.gauss.2015.nc"),
        v_wind_path=str(data_root / "vwnd.10m.gauss.2015.nc"),
        wind_preprocessor=preprocess_wind,
        windage_coeff=0.005,  # fraction of windspeed transferred to particle
        eddy_diffusivity=200,  # m^2 / s
        advection_start_date=ADVECTION_START,
        timestep=timedelta(hours=1),
        num_timesteps=24 * (ADVECTION_END - ADVECTION_START).days,
        save_period=4,
        overwrite_existing_files=True,
    )

    for path in out_paths:
        print("Animating trajectories...")
        animate_ocean_advection(outputfile_path=path, save=False)
        print("Plotting trajectories...")
        plot_ocean_trajectories(outputfile_path=path)


if __name__ == "__main__":
    main()



"""
This is the ADVECTOR 2D entry-point.
To use, add this repo's "src" directory to the python path, import this file, and execute.  E.g.
    import sys
    sys.path.append("<path_to_repo>/src")
    from run_advector_2D import run_advector_2D
    run_advector_2D(...)
See examples/ECCO_advect_2D.py for an example usage.
See docstring below for descriptions of arguments.
See src/forcing_data_specifications.md for detailed description of forcing data requirements.
See src/sourcefile_specifications.md for detailed description of sourcefile requirements.
See src/outputfile_specifications.md for detailed description of the outputfile created by this program.
"""

import datetime
from pathlib import Path
from typing import Callable, Optional, Tuple, List, Union

import xarray as xr
from dask.diagnostics import ProgressBar

from .drivers.chunked_kernel_driver import execute_chunked_kernel_computation
from .enums.advection_scheme import AdvectionScheme
from .enums.forcings import Forcing
from .io_tools.OutputWriter import OutputWriter2D
from .io_tools.open_sourcefiles import open_2d_sourcefiles
from .io_tools.open_vectorfiles import *
from .kernel_wrappers.Kernel2D import Kernel2D, Kernel2DConfig


def run_advector_2D(
    sourcefile_path: str,
    output_directory: str,
    u_water_path: Union[List[str], str],
    v_water_path: Union[List[str], str],
    advection_start_date: datetime.datetime,
    timestep: datetime.timedelta,
    num_timesteps: int,
    eddy_diffusivity: float = 0,
    advection_scheme: str = "taylor2",
    save_period: int = 1,
    opencl_device: Tuple[int, ...] = None,
    memory_utilization: float = 0.4,
    u_wind_path: Optional[Union[List[str], str]] = None,
    v_wind_path: Optional[Union[List[str], str]] = None,
    windage_coeff: Optional[float] = None,
    show_progress_bar: bool = True,
    water_preprocessor: Optional[Callable[[xr.Dataset], xr.Dataset]] = None,
    wind_preprocessor: Optional[Callable[[xr.Dataset], xr.Dataset]] = None,
    sourcefile_preprocessor: Optional[Callable[[xr.Dataset], xr.Dataset]] = None,
    overwrite_existing_files: bool = False,
) -> List[str]:
    """
    :param sourcefile_path: path to the particle sourcefile netcdf file.
        Can be a wildcard path as long as the individual sourcefiles can be properly concatenated along particle axis.
        See data_specifications.md for data requirements.
    :param output_directory: directory which will be populated with the outfiles.
        Existing files in this directory may be overwritten.
        See data_specifications.md for outputfile format details.
    :param u_water_path: wildcard path to the zonal current files.
        See data_specifications.md for data requirements.
    :param v_water_path: wildcard path to the meridional current files; see 'u_water_path'.
    :param advection_start_date: python datetime object denoting the start of the advection timeseries.
        Any particles which are scheduled to be released prior to this date will be released at this date.
    :param timestep: python timedelta object denoting the duration of each advection timestep.
    :param num_timesteps: length of the advection timeseries.
    :param eddy_diffusivity: (m^2 / s) controls the scale of each particle's random walk.  0 (default) has no effect.
        Note: since eddy diffusivity parameterizes ocean mechanics at smaller scales than the current files resolve,
            the value chosen should reflect the resolution of the current files.  Further, though eddy diffusivity in
            the real ocean varies widely in space and time, ADVECTOR uses one value everywhere, and the value should be
            selected with this in mind.
    :param advection_scheme: one of {"taylor2", "eulerian"}.
        "taylor2" is a second-order advection scheme as described in Black/Gay 1990 which improves adherence to circular
            streamlines compared to a first-order scheme.  This is the default.
        "eulerian" is the forward Euler method.
    :param save_period: controls how often to write output: particle state will be saved every {save_period} timesteps.
        For example, with timestep=one hour, and save_period=24, the particle state will be saved once per day.
    :param opencl_device: specifies hardware for computation.  If None (default), the user will receive a series of
        prompts which guides them through selecting a compute device.  To bypass this prompt, you can encode your
        answers to each of the prompts in a tuple, e.g. (0, 2).
    :param memory_utilization: this defines what percentage of the opencl device memory will be assumed usable.
        If you are using the CPU as your opencl device, don't set this above .5, because the CPU memory is shared by the
        host code (i.e. python-side), among other things.
        If you are using a dedicated compute device, this can be set close to 1.
        For example, on a dedicated compute device (e.g. GPU) with 2GB memory, setting memory_utilization = .95
            will allow the program to send a maximum of 1.9GB to the GPU at once.
        In general, set this as high as you can without running out of memory.
    :param u_wind_path: wildcard path to zonal 10-meter wind files; see 'u_water_path'.
        Wind is optional.  Simply omit this argument in order to disable drift due to wind.
    :param v_wind_path: wildcard path to meridional 10-meter wind files; see 'u_wind_path'.
    :param windage_coeff: fraction of wind speed that is transferred to particle.
        If u_wind_path is specified, i.e., wind is enabled, this value must be specified.
        Note: this value has a profound impact on results.
    :param show_progress_bar: whether to show progress bars for dask operations
    :param water_preprocessor: function to manipulate the water data just after loading.
        After preprocessor is applied, data must be compliant with forcing_data_specifications.md
    :param wind_preprocessor: see water_preprocessor
    :param sourcefile_preprocessor: see water_preprocessor, compliance info in sourcefile_specifications.md
    :param overwrite_existing_files: flag to skip warning prompts and clobber existing files,
        useful for running model with no possibility of user input
    :return: list of paths to the outputfiles
    """
    if show_progress_bar:
        ProgressBar(minimum=1).register()
    arguments = locals()
    try:
        scheme_enum = AdvectionScheme[advection_scheme]
    except KeyError:
        raise ValueError(
            f"Invalid argument advection_scheme; must be one of "
            f"{set(scheme.name for scheme in AdvectionScheme)}."
        )

    print("---INITIALIZING DATASETS---")
    print("Opening Sourcefiles...")
    p0 = open_2d_sourcefiles(
        sourcefile_path=sourcefile_path, preprocessor=sourcefile_preprocessor
    )

    forcing_data = {}
    print("Initializing Ocean Current...")
    forcing_data[Forcing.current] = open_2d_currents(
        u_path=u_water_path, v_path=v_water_path, preprocessor=water_preprocessor
    )

    if u_wind_path is not None and v_wind_path is not None:
        print("Initializing Wind...")
        forcing_data[Forcing.wind] = open_wind(
            u_path=u_wind_path, v_path=v_wind_path, preprocessor=wind_preprocessor
        )

    output_writer = OutputWriter2D(
        out_dir=Path(output_directory),
        basename="ADVECTOR_2D_output",
        sourcefile=p0,
        forcing_data=forcing_data,
        api_entry="src/run_advector_2D.py::run_advector_2D",
        api_arguments=arguments,
        overwrite_existing_files=overwrite_existing_files,
    )

    print("---COMMENCING ADVECTION---")
    out_paths = execute_chunked_kernel_computation(
        forcing_data=forcing_data,
        kernel_cls=Kernel2D,
        kernel_config=Kernel2DConfig(
            advection_scheme=scheme_enum,
            windage_coefficient=windage_coeff,
            eddy_diffusivity=eddy_diffusivity,
        ),
        output_writer=output_writer,
        p0=p0,
        start_time=advection_start_date,
        dt=timestep,
        num_timesteps=num_timesteps,
        save_every=save_period,
        platform_and_device=opencl_device,
        memory_utilization=memory_utilization,
    )

    return [str(p) for p in out_paths]



import subprocess
from pathlib import Path

import xarray as xr


def download_and_process_density(out_path: Path, user: str, password: str):
    temp_dir = out_path.parent / "density_temp"
    temp_dir.mkdir(exist_ok=True)
    url = (
        f"https://ecco.jpl.nasa.gov/drive/files/Version4/Release4/interp_monthly/"
        f"RHOAnoma/2015/RHOAnoma_2015_01.nc"
    )
    tmp_out_path = temp_dir / url.split("/")[-1]
    if not out_path.exists():
        subprocess.run(
            [
                "wget",
                "-q",
                "--show-progress",
                "--user",
                user,
                "--password",
                password,
                "-O",
                tmp_out_path,
                url,
            ]
        )

    rhoConst = 1029  # source: https://ecco.jpl.nasa.gov/drive/files/Version4/Release4/nctiles_monthly/README
    rho_anom = (
        xr.open_dataset(tmp_out_path)
        .swap_dims({"k": "Z", "j": "latitude", "i": "longitude"})
        .rename({"Z": "depth", "latitude": "lat", "longitude": "lon"})
        .drop_vars(("i", "j", "k", "timestep", "time_bnds"))
    )
    rho_anom = rho_anom.where(rho_anom != 0)  # zeros indicate no data; replace with nan
    rho_abs = (rhoConst + rho_anom).rename({"RHOAnoma": "rho"})
    rho_abs["rho"].attrs["long_name"] = "Seawater Density"
    rho_abs["rho"].attrs["units"] = "kg m^-3"

    rho_abs.to_netcdf(out_path)

    # clean up temp files
    tmp_out_path.unlink()
    temp_dir.rmdir()



import datetime
import gc
import logging
import time
from pathlib import Path
from typing import Dict, Tuple, List, Type

import numpy as np
import pandas as pd
import pyopencl as cl
import xarray as xr
from tqdm import tqdm

from .advection_chunking import chunk_advection_params
from ..enums.forcings import Forcing
from ..io_tools.OutputWriter import OutputWriter
from ..kernel_wrappers.Kernel import Kernel, KernelConfig
from ..kernel_wrappers.kernel_constants import EXIT_CODES


def execute_chunked_kernel_computation(
    forcing_data: Dict[Forcing, xr.Dataset],
    kernel_cls: Type[Kernel],
    kernel_config: KernelConfig,
    output_writer: OutputWriter,
    p0: xr.Dataset,
    start_time: datetime.datetime,
    dt: datetime.timedelta,
    num_timesteps: int,
    save_every: int,
    memory_utilization: float,
    platform_and_device: Tuple[int] = None,
) -> List[Path]:
    """
    Splits an advection computation into chunks, based on memory constraints of compute device,
    and executes the kernel over each chunk.
    :param forcing_data: dictionary holding whatever forcing data is available.
        valid keys: {"current", "wind", "seawater_density"}
    :param kernel_cls: kernel class used to execute the chunks
    :param kernel_config: dictionary to be passed to kernel_cls.__init__ as arg "config."
        See kernel_cls implementation for details on what configuration parameters are required.
    :param output_writer: object which is responsible for persisting the model output to disk
    :param p0: xarray Dataset storing particle initial state from sourcefile
    :param start_time: advection start time
    :param dt: timestep duration
    :param num_timesteps: number of timesteps
    :param save_every: how many timesteps between saving state.  Must divide num_timesteps.
    :param memory_utilization: fraction of the opencl device memory available for buffers
    :param platform_and_device: indices of platform/device to execute program.  None initiates interactive mode.
    :return: list of outputfile paths
    """
    num_particles = len(p0.p_id)
    advect_time = pd.date_range(start=start_time, freq=dt, periods=num_timesteps)
    # choose the device/platform we're running on
    if platform_and_device is None:
        context = cl.create_some_context(interactive=True)
    else:
        context = cl.create_some_context(answers=list(platform_and_device))

    # get the minimum RAM available on the specified compute devices.
    print("Chunking Datasets...")
    available_RAM = (
        min(device.global_mem_size for device in context.devices) * memory_utilization
    )
    advect_time_chunks, forcing_data_chunks = chunk_advection_params(
        device_bytes=available_RAM,
        forcing_data=forcing_data,
        num_particles=num_particles,
        advect_time=advect_time,
        save_every=save_every,
    )

    create_logger(output_writer.folder_path / "warnings.log")
    p0_chunk = p0.assign({"exit_code": ("p_id", np.zeros(len(p0.p_id)))})
    for i in tqdm(
        range(len(advect_time_chunks)),
        desc="PROGRESS",
        unit="chunk",
    ):
        print("")  # newline after the progress bar
        print(
            f"Advecting from {advect_time_chunks[i][0]} to {advect_time_chunks[i][-1]}..."
        )
        # create the kernel wrapper object, pass it arguments
        print("\tInitializing Kernel...")
        kernel = kernel_cls(
            forcing_data=forcing_data_chunks[i],
            p0=p0_chunk,
            config=kernel_config,
            advect_time=advect_time_chunks[i],
            save_every=save_every,
            context=context,
        )

        print("\tTriggering kernel execution...")
        P_chunk = kernel.execute()
        handle_errors(chunk=P_chunk, chunk_num=i + 1)
        data_loading_time = kernel.get_data_loading_time()
        buffer_time = kernel.get_buffer_transfer_time()
        execution_time = kernel.get_kernel_execution_time()
        memory_usage = kernel.get_memory_footprint()

        del kernel  # important for releasing memory for the next iteration
        gc.collect()

        print("\tWriting output to disk...")
        output_start = time.time()
        output_writer.write_output_chunk(P_chunk)
        output_time = time.time() - output_start

        p0_chunk = convert_final_state_to_initial_state(
            execution_result=P_chunk,
            advect_time=advect_time_chunks[i],
            previous_initial_state=p0_chunk,
        )

        print("\t----------BUFFER SIZES-----------")
        for key, value in memory_usage.items():
            print(f'\t{key+":":<20}{value / 1e6:10.3f} MB')
        print(f'\t{"Total:":<20}{sum(memory_usage.values()) / 1e6:10.3f} MB')
        print("\t---------EXECUTION TIME----------")
        print(f"\tData Loading:         {data_loading_time:10.3f}s")
        print(f"\tBuffer Read/Write:    {buffer_time:10.3f}s")
        print(f"\tKernel Execution:     {execution_time:10.3f}s")
        print(f"\tOutput Writing:       {output_time:10.3f}s")

    return output_writer.paths


def convert_final_state_to_initial_state(
    execution_result: xr.Dataset,
    previous_initial_state: xr.Dataset,
    advect_time: pd.DatetimeIndex,
) -> xr.Dataset:
    """Takes the final timestep of a computation, and convert it into an initial state to send to another kernel"""
    final_state = execution_result.isel(time=-1).copy(
        deep=True
    )  # copy in order to avoid mutating execution result
    # problem is, this ^ has nans for location of all the unreleased particles.  Restore that information here
    unreleased = final_state.release_date > advect_time[-2]
    for var in execution_result.data_vars:
        if "time" in execution_result[var].dims:
            final_state[var].loc[unreleased] = previous_initial_state[var].loc[
                unreleased
            ]
    return final_state


def create_logger(log_path: Path):
    """this sets up logging such that logs with level WARNING go to log_path,
    logs with level ERROR or greater go to log_path and stdout."""
    logging.basicConfig(
        filename=str(log_path),
        filemode="w",
        level=logging.WARNING,
        format="%(asctime)s %(message)s",
    )
    console = logging.StreamHandler()
    console.setLevel(logging.ERROR)
    logging.getLogger("").addHandler(console)


def handle_errors(chunk: xr.Dataset, chunk_num: int):
    reported_codes = np.unique(chunk.exit_code)
    if all(reported_codes == 0):
        return

    bad_codes = reported_codes[reported_codes != 0]
    logging.error(
        f"Error: {np.count_nonzero(chunk.exit_code)} particle(s) did not exit successfully: "
        f"exit code(s) {[f'{code} ({EXIT_CODES[code]})' for code in bad_codes]}"
    )
    for code in bad_codes:
        bad_ids = chunk.p_id[chunk.exit_code == code]
        logging.warning(
            f"Chunk {chunk_num}: {len(bad_ids)} particles exited with code {code}.  IDs: {list(bad_ids.values)}"
        )

    fatal_codes = bad_codes[bad_codes < 0]
    if fatal_codes:
        raise ValueError(
            f"Fatal error encountered, error code(s) "
            f"{[f'{code} ({EXIT_CODES[code]})' for code in fatal_codes]}; aborting"
        )



import time
from dataclasses import dataclass
from pathlib import Path
from typing import Dict

import numpy as np
import pandas as pd
import pyopencl as cl
import xarray as xr

from ..enums.advection_scheme import AdvectionScheme
from ..enums.forcings import Forcing
from ..kernel_wrappers.Field3D import Field3D, create_empty_2d_field, buffer_from_array
from ..kernel_wrappers.Kernel import Kernel, KernelConfig


KERNEL2D_SOURCE = open(Path(__file__).parent.parent / "kernels/kernel_2d.cl").read()


@dataclass
class Kernel2DConfig(KernelConfig):
    """Configuration for 2D Kernel.
    advection_scheme: which mathematical scheme to use for an advection step
    windage_coefficient: fraction of wind speed that is transferred to particles during advection
    eddy_diffusivity: (m^2 / s) controls the scale of each particle's Wiener-process random walk.
    """

    advection_scheme: AdvectionScheme
    windage_coefficient: float
    eddy_diffusivity: float


class Kernel2D(Kernel):
    """wrapper for src/kernels/kernel_2d.cl"""

    def __init__(
        self,
        forcing_data: Dict[Forcing, xr.Dataset],
        p0: xr.Dataset,
        advect_time: pd.DatetimeIndex,
        save_every: int,
        config: Kernel2DConfig,
        context: cl.Context,
    ):
        """
        :param forcing_data:
            required keys: {Forcing.current}
            optional keys: {Forcing.wind}
        :param p0: initial state of particles
        :param advect_time: the timeseries which the kernel advects on
        :param save_every: number of timesteps between each writing of particle state
        :param config: see Kernel2DConfig definition for details
        :param context: PyopenCL context for executing OpenCL programs
        """
        super().__init__(
            forcing_data=forcing_data,
            p0=p0,
            advect_time=advect_time,
            save_every=save_every,
            config=config,
            context=context,
        )

        # ---KERNEL ARGUMENT INITIALIZATION--- #
        # 1-to-1 for arguments in kernel_3d.cl::advect; see comments there for details
        print("\t\tLoading Current Data...")
        data_loading_start = time.time()
        self.current = Field3D(ds=forcing_data[Forcing.current], varnames=["U", "V"])

        print("\t\tLoading Wind Data...")
        if Forcing.wind in forcing_data:
            self.wind = Field3D(ds=forcing_data[Forcing.wind], varnames=["U", "V"])
        else:  # Windage disabled; pass a dummy field with singleton dimensions
            self.wind = create_empty_2d_field()
            # to flag the kernel that windage is disabled
            self.windage_coefficient = np.nan
        self.data_load_time = time.time() - data_loading_start

        # particle initialization
        self.x0 = p0.lon.values.astype(np.float32)
        self.y0 = p0.lat.values.astype(np.float32)
        self.release_date = p0.release_date.values.astype("datetime64[s]").astype(
            np.float64
        )

        # output_vectors (initialized to nan)
        out_shape = len(p0.lon) * len(self.out_time)
        self.X_out = np.full(out_shape, np.nan, dtype=np.float32)
        self.Y_out = np.full(out_shape, np.nan, dtype=np.float32)

        # physics
        self.config = config

        # debugging
        self.exit_code = p0.exit_code.values.astype(np.byte)

    def execute(self) -> xr.Dataset:
        """tranfers arguments to the compute device, triggers execution, returns result"""
        # perform argument check
        self._check_args()

        # write arguments to compute device
        print("\t\tWriting buffers to compute device...")
        write_start = time.time()
        current_args = self.current.create_kernel_arguments(context=self.context)
        wind_args = self.wind.create_kernel_arguments(context=self.context)
        p0_buffers = (
            buffer_from_array(arr=arr, context=self.context)
            for arr in (
                self.x0,
                self.y0,
                self.release_date,
            )
        )
        out_arrays = (self.X_out, self.Y_out, self.exit_code)
        out_buffers = [
            cl.Buffer(
                self.context,
                cl.mem_flags.READ_WRITE | cl.mem_flags.COPY_HOST_PTR,
                hostbuf=arr,
            )
            for arr in out_arrays
        ]
        self.buf_time = time.time() - write_start

        # execute the program
        print("\t\tExecuting kernel...")
        execution_start = time.time()
        self.cl_kernel.advect(
            self.queue,
            (len(self.x0),),
            None,
            *current_args,
            *wind_args,
            *p0_buffers,
            np.uint32(self.config.advection_scheme.value),
            np.float64(self.config.windage_coefficient),
            np.float64(self.config.eddy_diffusivity),
            self.start_time,
            self.dt,
            self.ntimesteps,
            self.save_every,
            *out_buffers,
        )
        # wait for the computation to complete
        self.queue.finish()
        self.kernel_time = time.time() - execution_start

        # Read back the results from the compute device
        print("\t\tCopying results back to host device...")
        read_start = time.time()
        for arr, buffer in zip(out_arrays, out_buffers):
            cl.enqueue_copy(self.queue, arr, buffer)
        self.buf_time += time.time() - read_start

        # create output dataset on top of initial state
        P = self.p0.copy(deep=True).assign_coords({"time": self.out_time})
        return P.assign(
            {
                "lon": (["p_id", "time"], self.X_out.reshape([len(P.p_id), -1])),
                "lat": (["p_id", "time"], self.Y_out.reshape([len(P.p_id), -1])),
                "exit_code": (["p_id"], self.exit_code),
            }
        )

    def get_memory_footprint(self) -> dict:
        return {
            "current": self.current.memory_usage_bytes(),
            "wind": self.wind.memory_usage_bytes(),
            "particles": (
                self.x0.nbytes
                + self.y0.nbytes
                + self.release_date.nbytes
                + self.X_out.nbytes
                + self.Y_out.nbytes
                + self.exit_code.nbytes
            ),
        }

    def get_data_loading_time(self) -> float:
        return self.data_load_time

    def get_buffer_transfer_time(self) -> float:
        return self.buf_time

    def get_kernel_execution_time(self) -> float:
        return self.kernel_time

    def _check_args(self):
        """ensure kernel arguments satisfy constraints"""

        # check particle positions valid
        assert np.nanmax(self.x0) < 180
        assert np.nanmin(self.x0) >= -180
        assert np.nanmax(self.y0) <= 90
        assert np.nanmin(self.y0) >= -90

        # check enum valid
        assert self.config.advection_scheme.value in (0, 1)

    @property
    def _kernel_source_code(self) -> str:
        return KERNEL2D_SOURCE



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Dict, List

import netCDF4
import numpy as np
import xarray as xr

import ADVECTOR
from ..enums.forcings import Forcing
from ..kernel_wrappers.kernel_constants import EXIT_CODES

SOURCEFILE_GROUP_NAME = "sourcefile"
MODEL_DOMAIN_GROUP_NAME = "model_domain"
CONFIGFILE_GROUP_NAME = "configfile"


class OutputWriter(ABC):
    def __init__(
        self,
        out_dir: Path,
        basename: str,
        sourcefile: xr.Dataset,
        forcing_data: Dict[Forcing, xr.Dataset],
        api_entry: str,
        api_arguments: dict,
        overwrite_existing_files: bool,
    ):
        """
        :param out_dir: directory to save outputfiles
        :param basename: base name of each outputfile (e.g. out_name = "3d_output" --> "3d_output_1993.nc")
        :param sourcefile: sourcefile, to be copied to outputfiles.
        :param forcing_data: xr.Datasets containing forcing datasets (e.g. currents, wind...)
        :param api_entry: the function where this is being called from (for traceability)
        :param api_arguments: dictionary containing info on the top-level API call
        """
        if out_dir.exists() and any(out_dir.iterdir()):
            print(
                f"DANGER: There are already files in '{out_dir}'! Contents may be overwritten!"
            )
            answer = "" if not overwrite_existing_files else "y"
            while answer not in {"y", "n"}:
                answer = input("Continue anyway? [y/n]: ")
            if answer == "n":
                raise SystemExit
        out_dir.mkdir(exist_ok=True, parents=True)

        self.folder_path = out_dir
        self.basename = basename
        self.current_year = None
        self.paths = []

        self.sourcefile = sourcefile
        self.forcing_meta = {
            forcing: xr.Dataset(ds.coords, attrs=ds.attrs)
            for forcing, ds in forcing_data.items()
        }
        self.api_entry = api_entry
        self.api_arguments = api_arguments

        self.model_domain = self._get_ocean_domain(forcing_data)

    @property
    @abstractmethod
    def _dataset_title(self) -> str:
        pass

    @property
    @abstractmethod
    def _group_names(self) -> List[str]:
        pass

    @abstractmethod
    def _get_ocean_domain(self, forcing_data: Dict[Forcing, xr.Dataset]) -> xr.Dataset:
        pass

    def _set_current_year(self, year: int):
        self.current_year = year
        self.paths.append(self.folder_path / f"{self.basename}_{year}.nc")

    def write_output_chunk(self, chunk: xr.Dataset):
        beginning_year = chunk.time.dt.year.values[0]
        end_year = chunk.time.dt.year.values[-1]

        for year in range(beginning_year, end_year + 1):
            chunk_year = chunk.isel({"time": chunk.time.dt.year == year})
            if year != self.current_year:
                self._set_current_year(year)
                self._write_first_chunk(chunk_year)
                self._copy_unexpected_variables(chunk_year)
            else:
                self._append_chunk(chunk_year)

    def _write_first_chunk(self, chunk: xr.Dataset):
        with netCDF4.Dataset(self.paths[-1], mode="w") as ds:
            # --- INITIALIZE PARTICLE TRAJECTORIES IN ROOT GROUP --- #
            ds.title = self._dataset_title
            ds.description = (
                "This file's root group contains timeseries location data "
                "for particles run through ADVECTOR.  This file also contains "
                "several other self-describing groups: "
                f"{self._group_names}."
            )
            ds.institution = "The Ocean Cleanup"
            ds.source = f"ADVECTOR Version {ADVECTOR.__version__}"
            ds.arguments = (
                f"The arguments of the call to {self.api_entry} which produced this "
                f"file are: {str(self.api_arguments)}"
            )

            ds.createDimension("p_id", len(chunk.p_id))
            ds.createDimension("time", None)  # unlimited dimension

            # Variables along only the static dimension, p_id
            p_id = ds.createVariable("p_id", chunk.p_id.dtype, ("p_id",))
            p_id[:] = chunk.p_id.values

            release_date = ds.createVariable("release_date", np.float64, ("p_id",))
            release_date.units = "seconds since 1970-01-01 00:00:00.0"
            release_date.calendar = "gregorian"
            release_date[:] = chunk.release_date.values.astype("datetime64[s]").astype(
                np.float64
            )

            exit_code = ds.createVariable("exit_code", np.byte, ("p_id",))
            exit_code.description = (
                "These codes are returned by the kernel when unexpected behavior occurs and the"
                "kernel must be terminated.  Their semantic meaning is provided in the "
                "'code_to_meaning' attribute of this variable."
            )
            exit_code.code_to_meaning = str(
                {code: meaning for code, meaning in EXIT_CODES.items() if code >= 0}
            )
            exit_code[:] = chunk.exit_code.values.astype(np.byte)

            # Variables that expand between chunks
            time = ds.createVariable("time", np.float64, ("time",))
            time.units = "seconds since 1970-01-01 00:00:00.0"
            time.calendar = "gregorian"
            time[:] = chunk.time.values.astype("datetime64[s]").astype(np.float64)

            lon = ds.createVariable("lon", chunk.lon.dtype, ("p_id", "time"))
            lon.units = "Degrees East"
            lon[:] = chunk.lon.values

            lat = ds.createVariable("lat", chunk.lat.dtype, ("p_id", "time"))
            lat.units = "Degrees North"
            lat[:] = chunk.lat.values

        # --- SAVE MODEL CONFIGURATION METADATA INTO GROUPS --- #
        self.model_domain.to_netcdf(self.paths[-1], mode="a", group="model_domain")
        self.sourcefile.to_netcdf(self.paths[-1], mode="a", group=SOURCEFILE_GROUP_NAME)
        for forcing, meta in self.forcing_meta.items():
            meta.attrs["group_description"] = (
                f"This group contains the coordinates of the fully concatenated {forcing.value} "
                "dataset, after it has been loaded into ADVECTOR, and global attributes "
                "from the first file in the dataset."
            )
            meta.to_netcdf(self.paths[-1], mode="a", group=forcing.name + "_meta")

    def _copy_unexpected_variables(self, chunk: xr.Dataset):
        """copy any variables along only p_id should be copied over as well"""
        with netCDF4.Dataset(self.paths[-1], mode="a") as ds:
            dims_except_p_id = [dim for dim in chunk.dims if dim != "p_id"]
            for varname in chunk.drop_dims(dims_except_p_id).variables:
                if varname not in ds.variables:
                    ds.createVariable(varname, chunk[varname].dtype, ("p_id",))
                    ds[varname].setncatts(chunk[varname].attrs)
                    ds[varname][:] = chunk[varname].values

    def _append_chunk(self, chunk: xr.Dataset):
        with netCDF4.Dataset(self.paths[-1], mode="a") as ds:
            time = ds.variables["time"]
            start_t = len(time)
            time[start_t:] = chunk.time.values.astype("datetime64[s]").astype(
                np.float64
            )

            lon = ds.variables["lon"]
            lon[:, start_t:] = chunk.lon.values

            lat = ds.variables["lat"]
            lat[:, start_t:] = chunk.lat.values

            exit_code = ds.variables["exit_code"]
            # overwrite with most recent codes; by design, nonzero codes cannot change
            exit_code[:] = chunk.exit_code.values


class OutputWriter2D(OutputWriter):
    def _get_ocean_domain(self, forcing_data: Dict[Forcing, xr.Dataset]) -> xr.Dataset:
        land_mask_varname = "land_mask"
        return (
            (forcing_data[Forcing.current]["U"].isel(time=0).isnull())
            .drop_vars(("time", "depth"), errors="ignore")
            .rename(land_mask_varname)
            .assign_attrs(
                {
                    "description": "boolean mask which identifies 'land', defined as "
                    "the null cells in the sea surface current velocity field."
                }
            )
            .to_dataset()
            .assign_attrs(
                {
                    "domain_definition": "The internal model domain is defined as all grid cells which are "
                    f"not land; land is defined by the variable '{land_mask_varname}'."
                }
            )
        )

    @property
    def _dataset_title(self) -> str:
        return "Trajectories of Floating Marine Debris"

    @property
    def _group_names(self) -> List[str]:
        return [MODEL_DOMAIN_GROUP_NAME, SOURCEFILE_GROUP_NAME] + list(
            forcing.name + "_meta" for forcing in self.forcing_meta.keys()
        )


class OutputWriter3D(OutputWriter):
    def __init__(
        self,
        out_dir: Path,
        basename: str,
        configfile: xr.Dataset,
        sourcefile: xr.Dataset,
        forcing_data: Dict[Forcing, xr.Dataset],
        api_entry: str,
        api_arguments: dict,
        overwrite_existing_files: bool,
    ):
        """
        :param configfile: configfile, to be copied to outputfiles
        see OutputWriter for other arg descriptions
        """
        super().__init__(
            out_dir=out_dir,
            basename=basename,
            sourcefile=sourcefile,
            forcing_data=forcing_data,
            api_entry=api_entry,
            api_arguments=api_arguments,
            overwrite_existing_files=overwrite_existing_files,
        )
        self.configfile = configfile

    def _write_first_chunk(self, chunk: xr.Dataset):
        super()._write_first_chunk(chunk=chunk)

        # --- SAVE MODEL CONFIGURATION METADATA INTO GROUPS --- #
        self.configfile.to_netcdf(self.paths[-1], mode="a", group=CONFIGFILE_GROUP_NAME)

        with netCDF4.Dataset(self.paths[-1], mode="a") as ds:
            # --- INITIALIZE PARTICLE TRAJECTORIES IN ROOT GROUP --- #
            radius = ds.createVariable("radius", np.float64, ("p_id",))
            radius.units = "meters"
            radius[:] = chunk.radius.values.astype(np.float64)

            density = ds.createVariable("density", np.float64, ("p_id",))
            density.units = "kg m^-3"
            density[:] = chunk.density.values.astype(np.float64)

            corey_shape_factor = ds.createVariable(
                "corey_shape_factor", np.float64, ("p_id",)
            )
            corey_shape_factor.units = "unitless"
            corey_shape_factor[:] = chunk.corey_shape_factor.values.astype(np.float64)

            depth = ds.createVariable("depth", chunk.depth.dtype, ("p_id", "time"))
            depth.units = "meters"
            depth.positive = "up"
            depth[:] = chunk.depth.values

    def _append_chunk(self, chunk: xr.Dataset):
        with netCDF4.Dataset(self.paths[-1], mode="a") as ds:
            start_t = len(ds.variables["time"])
        super()._append_chunk(chunk=chunk)
        with netCDF4.Dataset(self.paths[-1], mode="a") as ds:
            depth = ds.variables["depth"]
            depth[:, start_t:] = chunk.depth.values

    def _get_ocean_domain(self, forcing_data: Dict[Forcing, xr.Dataset]) -> xr.Dataset:
        return (
            forcing_data[Forcing.current]["bathymetry"]
            .assign_attrs(
                {
                    "description": "depth of ocean floor, defined as the bottom bound "
                    "of the first non-null cell of the ocean current "
                    "velocity field along the ascending depth dimension."
                }
            )
            .to_dataset()
            .assign_attrs(
                {
                    "domain_definition": "The internal model domain is defined as all points in "
                    "(lat, lon, depth) space where depth >= bathymetry, "
                    "excluding cells where bathymetry == 0."
                }
            )
        )

    @property
    def _dataset_title(self) -> str:
        return "Trajectories of Marine Debris"

    @property
    def _group_names(self) -> List[str]:
        return [
            MODEL_DOMAIN_GROUP_NAME,
            SOURCEFILE_GROUP_NAME,
            CONFIGFILE_GROUP_NAME,
        ] + list(forcing.name + "_meta" for forcing in self.forcing_meta.keys())



from enum import Enum


class Forcing(Enum):
    """use .name for variable name, .value for human readable name"""

    current = "current"
    wind = "10-meter wind"
    seawater_density = "seawater density"



from typing import Optional, Set, List, Union, Callable

import dask
import numpy as np
import xarray as xr

from ..io_tools.create_bathymetry import create_bathymetry_from_land_mask


def open_3d_currents(
    u_path: Union[List[str], str],
    v_path: Union[List[str], str],
    w_path: Union[List[str], str],
    preprocessor: Optional[Callable[[xr.Dataset], xr.Dataset]],
):
    """
    :param u_path: wildcard path to the zonal vector files.  Fed to glob.glob.
    :param v_path: wildcard path to the meridional vector files.
    :param w_path: wildcard path to the vertical vector files.
    :param preprocessor: func to call on the xarray dataset to perform operation before loading in advector, such as renaming variables.
    """
    currents = open_vectorfield(
        paths=[u_path, v_path, w_path],
        varnames={"U", "V", "W"},
        keep_depth_dim=True,
        preprocessor=preprocessor,
    )
    # encode the model domain, taken as where all the current components are non-null, as bathymetry
    print("Calculating bathymetry of current dataset...")
    first_timestep = currents.isel(time=0)  # only need one timestep
    land_mask = (
        first_timestep.U.isnull()
        | first_timestep.V.isnull()
        | first_timestep.W.isnull()
    )

    return xr.merge(
        (currents, create_bathymetry_from_land_mask(land_mask)),
        combine_attrs="override",
    )


def open_2d_currents(
    u_path: Union[List[str], str],
    v_path: Union[List[str], str],
    preprocessor: Optional[Callable[[xr.Dataset], xr.Dataset]],
):
    """
    :param u_path: wildcard path to the zonal vector files.  Fed to glob.glob.
    :param v_path: wildcard path to the meridional vector files.
    :param preprocessor: func to call on the xarray dataset to perform operation before loading in advector, such as renaming variables.
    """
    return open_vectorfield(
        paths=[u_path, v_path],
        varnames={"U", "V"},
        keep_depth_dim=False,
        preprocessor=preprocessor,
    )


def open_seawater_density(
    path: Union[List[str], str],
    preprocessor: Optional[Callable[[xr.Dataset], xr.Dataset]],
) -> xr.Dataset:
    """
    :param path: wildcard path to the seawater density files.  Fed to glob.glob.
    :param preprocessor: func to call on the xarray dataset to perform operation before loading in advector, such as renaming variables.
    """
    return open_vectorfield(
        paths=[path], varnames={"rho"}, keep_depth_dim=True, preprocessor=preprocessor
    )


def open_wind(
    u_path: Union[List[str], str],
    v_path: Union[List[str], str],
    preprocessor: Optional[Callable[[xr.Dataset], xr.Dataset]],
):
    """
    :param u_path: wildcard path to the zonal vector files.  Fed to glob.glob.
    :param v_path: wildcard path to the meridional vector files.
    :param preprocessor: func to call on the xarray dataset to perform operation before loading in advector, such as renaming variables.
    """
    return open_vectorfield(
        paths=[u_path, v_path],
        varnames={"U", "V"},
        keep_depth_dim=False,
        preprocessor=preprocessor,
    )


def open_vectorfield(
    paths: List[Union[List[str], str]],
    varnames: Set[str],
    keep_depth_dim: bool,
    preprocessor: Optional[Callable[[xr.Dataset], xr.Dataset]],
) -> xr.Dataset:
    print("\tOpening NetCDF files...")
    vectors = xr.merge(
        (
            xr.open_mfdataset(
                path,
                data_vars="minimal",
                parallel=True,
            )
            for path in paths
        ),
        combine_attrs="override",  # use first file's attributes
    )
    if preprocessor is not None:
        print("\tAppling preprocessor...")
        vectors = preprocessor(vectors)

    vectors = vectors[list(varnames)]  # drop any additional variables

    if keep_depth_dim:
        # convert positive-down depth to positive-up if necessary
        if np.all(vectors.depth >= 0):
            print("\tConverting depth to positive-up...")
            vectors["depth"] = -1 * vectors.depth
        if not np.all(np.diff(vectors.depth) >= 0):
            print("\tDepth dimension not sorted.  Sorting..")
            with dask.config.set(**{"array.slicing.split_large_chunks": False}):
                vectors = vectors.sortby("depth", ascending=True)
        expected_dims = {"lat", "lon", "time", "depth"}
    else:
        if "depth" in vectors.dims:
            print("\tExtracting nearest level to depth=0...")
            vectors = vectors.sel(depth=0, method="nearest")
        expected_dims = {"lat", "lon", "time"}
    assert (
        set(vectors.dims) == expected_dims
    ), f"Unexpected/missing dimension(s) (Found : {set(vectors.dims)}, Expected : {expected_dims})"

    if max(vectors.lon) > 180:
        print("\tRolling longitude domain from [0, 360) to [-180, 180).")
        print(
            "\tThis operation is expensive.  You may want to preprocess your data to the correct domain."
        )
        with dask.config.set(**{"array.slicing.split_large_chunks": True}):
            vectors["lon"] = ((vectors.lon + 180) % 360) - 180
            vectors = vectors.sortby("lon")

    return vectors



import glob
import subprocess
from pathlib import Path

import pandas as pd
import xarray as xr
from tqdm import tqdm


def download_and_process_currents(out_dir: Path):
    out_dir.mkdir(exist_ok=True)
    raw_dir = out_dir / "raw"
    raw_dir.mkdir(exist_ok=True)
    unmasked_dir = out_dir / "unmasked"
    unmasked_dir.mkdir(exist_ok=True)

    Z = None  # placeholder
    while True:
        none_failed = True
        for day in pd.date_range("2015-01-01", "2015-01-31", freq="D"):
            for varname in ("EVEL", "NVEL", "WVELMASS"):
                filename = f"{varname}_{day.strftime('%Y_%m_%d')}.nc"
                url = (
                    f"https://data.nas.nasa.gov/ecco/download_data.php?"
                    f"file=/eccodata/llc_90/ECCOv4/Release4/interp_daily/"
                    f"{varname}/2015/{day.dayofyear:03d}/{filename}"
                )
                unmasked_path = unmasked_dir / filename
                if unmasked_path.exists():
                    continue

                # download the raw data
                raw_out_path = raw_dir / filename
                subprocess.run(
                    [
                        "wget",
                        "-q",
                        "--show-progress",
                        "-O",
                        raw_out_path,
                        url,
                    ]
                )

                # open it up for some editing
                try:
                    ds = xr.open_dataset(raw_out_path)
                except (OSError, ValueError):
                    # file didn't download properly, delet and try again later
                    none_failed = False
                    raw_out_path.unlink()
                    continue

                # swap around some dims, interpolate W to same depth levels as U/V
                ds = ds.swap_dims({"j": "latitude", "i": "longitude"})
                if varname != "WVELMASS":
                    ds = ds.swap_dims({"k": "Z"}).drop_vars("k")
                else:
                    if Z is None:
                        Z = xr.open_dataset(unmasked_dir / "EVEL_2015_01_01.nc").Z
                    ds = (
                        ds.swap_dims({"k_l": "Zl"})
                        .interp(Zl=Z)
                        .drop_vars(("Zl", "k_l"))
                    )
                ds = ds.drop_vars(("i", "j", "timestep", "time_bnds"))
                ds.to_netcdf(unmasked_path)
                raw_out_path.unlink()

        if none_failed:
            print("All files downloaded.")
            raw_dir.rmdir()
            break

    # now we need to compute a land mask.  It isn't included (odd),
    # so all we can do is test where every single file is EXACTLY zero.
    print("Computing land mask...")
    files = list(sorted(Path(f) for f in glob.glob(str(unmasked_dir / "*VEL*.nc"))))
    currents = xr.open_mfdataset(files)
    land = (
        ((currents.NVEL == 0) & (currents.EVEL == 0) & (currents.WVELMASS == 0))
        .all(dim="time")
        .load()
    )
    # set the very bottom layer to land; EVEL and NVEL have data in this layer,
    # but WVELMASS has none.  We want to be consistent.
    land = land.where(land.Z != land.Z[-1], True)

    print("Appling land mask to files...")
    for file in tqdm(files):
        ds = xr.open_dataset(file)
        ds = ds.where(~land)
        ds.to_netcdf(out_dir / file.name)
        file.unlink()

    unmasked_dir.unlink()



import glob
from typing import Optional, Set, Callable, Optional
import xarray as xr


def open_3d_sourcefiles(
    sourcefile_path: str,
    preprocessor: Optional[Callable[[xr.Dataset], xr.Dataset]],
) -> xr.Dataset:
    return open_sourcefiles(
        sourcefile_path=sourcefile_path,
        preprocessor=preprocessor,
        expected_vars={
            "p_id",
            "lon",
            "lat",
            "depth",
            "radius",
            "density",
            "corey_shape_factor",
            "release_date",
        },
    )


def open_2d_sourcefiles(
    sourcefile_path: str,
    preprocessor: Optional[Callable[[xr.Dataset], xr.Dataset]],
) -> xr.Dataset:
    return open_sourcefiles(
        sourcefile_path=sourcefile_path,
        preprocessor=preprocessor,
        expected_vars={"p_id", "lon", "lat", "release_date"},
    )


def open_sourcefiles(
    sourcefile_path: str,
    preprocessor: Optional[Callable[[xr.Dataset], xr.Dataset]],
    expected_vars: Set[str],
) -> xr.Dataset:
    """
    :param sourcefile_path: path to the particle sourcefile netcdf file.  Absolute path safest, use relative paths with caution.
    :param expected_vars: variable names which are expected to be in sourcefile
    :param preprocessor: func to call on the xarray dataset to perform operation before loading in advector, such as renaming variables.
    """

    sourcefile = xr.open_mfdataset(
        sorted(glob.glob(sourcefile_path)),
        parallel=True,
        preprocess=preprocessor,
        combine="nested",
        concat_dim="p_id",
    )

    # make sure there's only one dimension
    dims = list(sourcefile.dims.keys())
    assert len(dims) == 1, "sourcefile has more than one dimension."

    for var in expected_vars:
        assert (
            var in sourcefile.variables
        ), f"missing variable '{var}'.  If differently named, pass in mapping."

    sourcefile.load()  # persist into RAM

    return sourcefile



import time
import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Optional

import numpy as np
import pandas as pd
import pyopencl as cl
import xarray as xr

from ..enums.advection_scheme import AdvectionScheme
from ..enums.forcings import Forcing
from ..kernel_wrappers.Field3D import (
    Field3D,
    create_empty_2d_field,
    is_sorted_ascending,
    buffer_from_array,
)
from ..kernel_wrappers.Kernel import Kernel, KernelConfig


KERNEL3D_SOURCE = open(Path(__file__).parent.parent / "kernels/kernel_3d.cl").read()


@dataclass
class Kernel3DConfig(KernelConfig):
    """Configuration for 3D Kernel.
    advection_scheme: which mathematical scheme to use for an advection step
    windage_multiplier: scales the physically-motivated windage contribution
    wind_mixing_enabled: toggles wind-driven mixing
    max_wave_height: (m) see configfile_specifications.md for details
    wave_mixing_depth_factor: see configfile_specifications.md for details
    eddy_diffusivity: dataset with vertical profiles of horizontal/vertical eddy diffusivity
        variables [m^2 / s] (dim [m]):
        horizontal_diffusivity (z_hd)
        vertical_diffusivity (z_vd)
    """

    advection_scheme: AdvectionScheme
    windage_multiplier: Optional[float]
    wind_mixing_enabled: bool
    max_wave_height: float
    wave_mixing_depth_factor: float
    eddy_diffusivity: xr.Dataset


class Kernel3D(Kernel):
    """wrapper for src/kernels/kernel_3d.cl"""

    def __init__(
        self,
        forcing_data: Dict[Forcing, xr.Dataset],
        p0: xr.Dataset,
        advect_time: pd.DatetimeIndex,
        save_every: int,
        config: Kernel3DConfig,
        context: cl.Context,
    ):
        """
        :param forcing_data:
            required keys: {Forcing.current, Forcing.seawater_density}
            optional keys: {Forcing.wind}
        :param p0: initial state of particles
        :param advect_time: the timeseries which the kernel advects on
        :param save_every: number of timesteps between each writing of particle state
        :param config: see Kernel3DConfig definition for details
        :param context: PyopenCL context for executing OpenCL programs
        """
        super().__init__(
            forcing_data=forcing_data,
            p0=p0,
            advect_time=advect_time,
            save_every=save_every,
            config=config,
            context=context,
        )

        # ---KERNEL ARGUMENT INITIALIZATION--- #
        # 1-to-1 for arguments in kernel_3d.cl::advect; see comments there for details
        print("\t\tLoading Current Data...")
        data_loading_start = time.time()
        self.current = Field3D(
            ds=forcing_data[Forcing.current], varnames=["U", "V", "W", "bathymetry"]
        )

        print("\t\tLoading Wind Data...")
        if "wind" in forcing_data:
            self.wind = Field3D(ds=forcing_data[Forcing.wind], varnames=["U", "V"])
        else:  # Windage disabled; pass a dummy field with singleton dimensions
            self.wind = create_empty_2d_field()
            self.windage_multiplier = (
                np.nan
            )  # to flag the kernel that windage is disabled

        print("\t\tLoading Seawater Density Data...")
        self.seawater_density = Field3D(
            ds=forcing_data[Forcing.seawater_density],
            varnames=["rho"],
            non_uniform_time=True,
        )
        self.data_load_time = time.time() - data_loading_start

        # particle initialization
        self.x0 = p0.lon.values.astype(np.float32)
        self.y0 = p0.lat.values.astype(np.float32)
        self.z0 = p0.depth.values.astype(np.float32)
        self.release_date = p0.release_date.values.astype("datetime64[s]").astype(
            np.float64
        )
        self.radius = p0.radius.values.astype(np.float64)
        self.density = p0.density.values.astype(np.float64)
        self.corey_shape_factor = p0.corey_shape_factor.values.astype(np.float64)

        # output_vectors (initialized to nan)
        out_shape = len(p0.lon) * len(self.out_time)
        self.X_out = np.full(out_shape, np.nan, dtype=np.float32)
        self.Y_out = np.full(out_shape, np.nan, dtype=np.float32)
        self.Z_out = np.full(out_shape, np.nan, dtype=np.float32)

        # physics
        self.config = config

        # eddy diffusivity
        self.horizontal_eddy_diffusivity_z = config.eddy_diffusivity.z_hd.values.astype(
            np.float64
        )
        self.horizontal_eddy_diffusivity_values = (
            config.eddy_diffusivity.horizontal_diffusivity.values.astype(np.float64)
        )
        self.vertical_eddy_diffusivity_z = config.eddy_diffusivity.z_vd.values.astype(
            np.float64
        )
        self.vertical_eddy_diffusivity_values = (
            config.eddy_diffusivity.vertical_diffusivity.values.astype(np.float64)
        )

        # debugging
        self.exit_code = p0.exit_code.values.astype(np.byte)

    def execute(self) -> xr.Dataset:
        """tranfers arguments to the compute device, triggers execution, returns result"""
        # perform argument check
        self._check_args()

        # write arguments to compute device
        print("\t\tWriting buffers to compute device...")
        write_start = time.time()
        current_args = self.current.create_kernel_arguments(context=self.context)
        wind_args = self.wind.create_kernel_arguments(context=self.context)
        seawater_density_args = self.seawater_density.create_kernel_arguments(
            context=self.context
        )
        p0_buffers = [
            buffer_from_array(arr=arr, context=self.context)
            for arr in (
                self.x0,
                self.y0,
                self.z0,
                self.release_date,
                self.radius,
                self.density,
                self.corey_shape_factor,
            )
        ]
        eddy_diffusity_buffers = [
            buffer_from_array(arr=arr, context=self.context)
            for arr in (
                self.horizontal_eddy_diffusivity_z,
                self.horizontal_eddy_diffusivity_values,
                self.vertical_eddy_diffusivity_z,
                self.vertical_eddy_diffusivity_values,
            )
        ]

        out_arrays = (self.X_out, self.Y_out, self.Z_out, self.exit_code)
        out_buffers = [
            cl.Buffer(
                self.context,
                cl.mem_flags.READ_WRITE | cl.mem_flags.COPY_HOST_PTR,
                hostbuf=arr,
            )
            for arr in out_arrays
        ]
        self.buf_time = time.time() - write_start

        # execute the program
        print("\t\tExecuting kernel...")
        execution_start = time.time()
        self.cl_kernel.advect(
            self.queue,
            (len(self.x0),),
            None,
            *current_args,
            *wind_args,
            *seawater_density_args,
            *p0_buffers,
            np.uint32(self.config.advection_scheme.value),
            np.float64(self.windage_multiplier),
            np.uint32(self.config.wind_mixing_enabled),
            np.float64(self.config.max_wave_height),
            np.float64(self.config.wave_mixing_depth_factor),
            *eddy_diffusity_buffers,
            np.uint32(len(self.horizontal_eddy_diffusivity_values)),
            np.uint32(len(self.vertical_eddy_diffusivity_values)),
            self.start_time,
            self.dt,
            self.ntimesteps,
            self.save_every,
            *out_buffers,
        )
        # wait for the computation to complete
        self.queue.finish()
        self.kernel_time = time.time() - execution_start

        # Read back the results from the compute device
        print("\t\tCopying results back to host device...")
        read_start = time.time()
        for arr, buffer in zip(out_arrays, out_buffers):
            cl.enqueue_copy(self.queue, arr, buffer)
        self.buf_time += time.time() - read_start

        # create output dataset on top of initial state
        P = self.p0.copy(deep=True).assign_coords({"time": self.out_time})
        return P.assign(
            {
                "lon": (["p_id", "time"], self.X_out.reshape([len(P.p_id), -1])),
                "lat": (["p_id", "time"], self.Y_out.reshape([len(P.p_id), -1])),
                "depth": (["p_id", "time"], self.Z_out.reshape([len(P.p_id), -1])),
                "exit_code": (["p_id"], self.exit_code),
            }
        )

    def get_memory_footprint(self) -> dict:
        return {
            "current": self.current.memory_usage_bytes(),
            "wind": self.wind.memory_usage_bytes(),
            "seawater_density": self.seawater_density.memory_usage_bytes(),
            "particles": (
                self.x0.nbytes
                + self.y0.nbytes
                + self.z0.nbytes
                + self.release_date.nbytes
                + self.density.nbytes
                + self.radius.nbytes
                + self.corey_shape_factor.nbytes
                + self.X_out.nbytes
                + self.Y_out.nbytes
                + self.Z_out.nbytes
                + self.exit_code.nbytes
            ),
        }

    def get_data_loading_time(self) -> float:
        return self.data_load_time

    def get_buffer_transfer_time(self) -> float:
        return self.buf_time

    def get_kernel_execution_time(self) -> float:
        return self.kernel_time

    def _check_args(self):
        """ensure kernel arguments satisfy constraints"""

        # check eddy diffusion valid
        assert is_sorted_ascending(self.horizontal_eddy_diffusivity_z)
        assert is_sorted_ascending(self.vertical_eddy_diffusivity_z)

        # check particle positions valid
        assert np.nanmax(self.x0) < 180
        assert np.nanmin(self.x0) >= -180
        assert np.nanmax(self.y0) <= 90
        assert np.nanmin(self.y0) >= -90
        assert np.nanmax(self.z0) <= 0

        # check corey shape factor valid
        assert np.all((0.15 < self.corey_shape_factor) & (self.corey_shape_factor <= 1))

        # check enum valid
        assert self.config.advection_scheme.value in (0, 1)

        # issue warning if wind timestep is smaller than one day
        if np.any(
            np.diff(self.wind.coords["t"]) < pd.Timedelta(days=1).total_seconds()
        ):
            warnings.warn(
                "Timestep of wind data is less than a day.  The kernel assumes a fully developed sea state from each "
                "wind datum; short timesteps mean this is a bad assumption.  Use wind data averaged over a longer "
                "timestep, or complain to the developers (or both)."
            )

    @property
    def _kernel_source_code(self) -> str:
        return KERNEL3D_SOURCE



import numpy as np
import xarray as xr


def create_bathymetry_from_land_mask(land_mask: xr.DataArray) -> xr.DataArray:
    """Method: identifies the lower depth bound of the shallowest
        ocean cell (non-null) in each vertical grid column.
    :param land_mask: dimensions {time, depth, lat, lon}, boloean array, True where cell is land"""
    assert np.all(land_mask.depth <= 0), "depth coordinate must be positive up"
    assert np.all(
        np.diff(land_mask.depth) > 0
    ), "depth coordinate must be sorted ascending"

    # In the kernel, particles look up data based on the nearest cell-center.
    # Thus cell bounds are the midpoints between each centers.
    # Very top cell bound is surface, and bottom cell bounds are
    # assumed to be symmetric about bottom cell center.
    depth_diff = np.diff(land_mask.depth)
    depth_bnds = np.concatenate(
        [
            land_mask.depth.values[:1] - depth_diff[0] / 2,
            land_mask.depth.values[:-1] + depth_diff / 2,
            [0],
        ]
    )

    bathy = (
        (~land_mask)
        .assign_coords({"depth": depth_bnds[:-1]})
        .idxmax(dim="depth")
        .where(~land_mask.isel(depth=-1), depth_bnds[-1])
    )

    bathy = bathy.drop(["time", "depth"])
    bathy.name = "bathymetry"
    bathy.attrs = {"units": "m", "positive": "up"}

    return bathy



from typing import Tuple

import xarray as xr

EDDY_DIFFUSIVITY_VARIABLES = {
    "z_hd",
    "horizontal_diffusivity",
    "z_vd",
    "vertical_diffusivity",
}
OPTIONAL_VARIABLES_WITH_DEFAULTS = {
    "max_wave_height": 20,  # meters
    "wave_mixing_depth_factor": 10,  # unitless
}


def unpack_configfile(configfile_path: str) -> Tuple[xr.Dataset, float, float]:
    """
    :param configfile_path: path to configfile
    return: (eddy diffusivity, max wave height, wave mixing depth factor)
    """
    configfile = xr.open_dataset(configfile_path)
    for var in EDDY_DIFFUSIVITY_VARIABLES:
        assert (
            var in configfile.variables
        ), f"configfile {configfile_path} missing variable '{var}'"
    eddy_diffusivity = configfile[list(EDDY_DIFFUSIVITY_VARIABLES)]
    eddy_diffusivity = eddy_diffusivity.sortby(["z_hd", "z_vd"], ascending=True)

    if "max_wave_height" in configfile.variables:
        max_wave_height = float(configfile["max_wave_height"])
    else:
        max_wave_height = OPTIONAL_VARIABLES_WITH_DEFAULTS["max_wave_height"]

    if "wave_mixing_depth_factor" in configfile.variables:
        wave_mixing_depth_factor = float(configfile["wave_mixing_depth_factor"])
    else:
        wave_mixing_depth_factor = OPTIONAL_VARIABLES_WITH_DEFAULTS[
            "wave_mixing_depth_factor"
        ]

    return eddy_diffusivity, max_wave_height, wave_mixing_depth_factor



import subprocess
import tempfile
from getpass import getpass


def get_ECCO_credentials():
    print(
        "Download requires authentication. "
        "You can find your WebDAV credenials and/or "
        "create an account here: https://ecco.jpl.nasa.gov/drive/"
    )
    test_url = "https://ecco.jpl.nasa.gov/drive/files/README"
    while True:
        user = input("Enter your WebDAV username: ")
        password = getpass("Enter your WebDAV password: ")
        with tempfile.TemporaryDirectory() as tmp_dir:
            wget_command = (
                f"wget --user {user} --password {password} -P {tmp_dir} {test_url}"
            )
            response = subprocess.getoutput(wget_command)

        if "200 OK" in response:
            print("Authentication Successful.")
            return user, password
        elif response.count("401 Unauthorized") == 2:
            # always 1 count because of the way the server handles
            # the request.  Second count means the user/pass is bad.
            print("Authentication failed.  Try again.")
            print("Find your WebDAV credenials here: https://ecco.jpl.nasa.gov/drive/")
        else:
            raise RuntimeError(
                "Unexpected failure while authenticating credentials.  Aborting.\n"
                f"Test wget command: '{wget_command}'\n"
                f"Output from test wget command:\n '{response}'"
            )
